// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
// vi: set et ts=4 sw=4 sts=4:
/*****************************************************************************
 *   See the file COPYING for full copying permissions.                      *
 *                                                                           *
 *   This program is free software: you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation, either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
 *****************************************************************************/
#ifndef DUMUX_MCWHORTERPROBLEM_HH
#define DUMUX_MCWHORTERPROBLEM_HH

#include <dumux/common/properties.hh>
#include <dumux/common/parameters.hh>
#include <dumux/porousmediumflow/2p/sequential/impes/problem.hh>

#include "../buckleyleverett/pseudooil.hh"
#include "../buckleyleverett/pseudoh2o.hh"
#include "mcwhorteranalyticsolution.hh"
#include <lecture/common/viplaboutput.hh>

namespace Dumux {

/*!
 * \brief McWhorter diffusion problem
 */
template<class TypeTag>
class McWhorterProblem: public IMPESProblem2P<TypeTag>
{
    using ParentType = IMPESProblem2P<TypeTag>;
    using GridView = typename GetPropType<TypeTag, Properties::GridGeometry>::GridView;
    using Scalar = GetPropType<TypeTag, Properties::Scalar>;
    using TimeManager = GetPropType<TypeTag, Properties::TimeManager>;
    using FluidSystem = GetPropType<TypeTag, Properties::FluidSystem>;
    using FluidState = GetPropType<TypeTag, Properties::FluidState>;
    using CellData = GetPropType<TypeTag, Properties::CellData>;
    using BoundaryTypes = GetPropType<TypeTag, Properties::SequentialBoundaryTypes>;
    using PrimaryVariables = typename GetProp<TypeTag, Properties::SolutionTypes>::PrimaryVariables;
    using Indices = GetPropType<TypeTag, Properties::Indices>;
    static constexpr int dim = GridView::dimension;
    static constexpr int dimWorld = GridView::dimensionworld;
    static constexpr int wPhaseIdx = Indices::wPhaseIdx;
    static constexpr int nPhaseIdx = Indices::nPhaseIdx;
    static constexpr int pnIdx = Indices::pnIdx;
    static constexpr int swIdx = Indices::swIdx;
    static constexpr int pressEqIdx = Indices::pressureEqIdx;
    static constexpr int satEqIdx = Indices::satEqIdx;
    using GlobalPosition = Dune::FieldVector<Scalar, dimWorld>;
    using Element = typename GridView::Traits::template Codim<0>::Entity;
    using Intersection = typename GridView::Intersection;

public:
    /*!
     * \brief DOC ME!
     * \param timeManager DOC ME!
     * \param gridView DOC ME!
     */
    McWhorterProblem(TimeManager& timeManager, typename GridView::Grid &grid)
    : ParentType(timeManager, grid)
    , eps_(1e-8)
    , pLeftBc_ ( 1.0e5 )
    , analyticSolution_(*this)
    , viplabOutput_(*this)
    {
        this->setOutputInterval( getParam<int>("Output.Interval") );

        PseudoOil<Scalar>::setViscosity( getParam<Scalar>("Fluid.ViscosityW") );
        PseudoH2O<Scalar>::setViscosity( getParam<Scalar>("Fluid.ViscosityNW") );

        PseudoOil<Scalar>::setDensity( getParam<Scalar>("Fluid.DensityW") );
        PseudoH2O<Scalar>::setDensity( getParam<Scalar>("Fluid.DensityNW") );

        swr_ = getParam<Scalar>("SpatialParams.Swr");
        snr_ = getParam<Scalar>("SpatialParams.Snr");

        paraviewOutput_ = getParam<bool>("Output.paraviewOutput", true);
    }

    void init()
    {
        ParentType::init();
        analyticSolution_.initialize();
    }

    /*!
     * \brief The problem name.
     *
     * This is used as a prefix for files generated by the simulation.
     */
    const std::string name() const
    {
        return "McWhorter";
    }

    bool shouldWriteRestartFile() const
    {
        return false;
    }

    void postTimeStep()
    {
        analyticSolution_.calculateAnalyticSolution();
        ParentType::postTimeStep();
    };

    void addOutputVtkFields()
    {
        ParentType::addOutputVtkFields();
        analyticSolution_.addOutputVtkFields(this->resultWriter());
    }

    /*!
     * \brief Returns the temperature within the domain.
     *
     * This problem assumes a temperature of 10 degrees Celsius.
     * \param globalPos DOC ME!
     */
    Scalar temperatureAtPos(const GlobalPosition& globalPos) const
    {
        return 273.15 + 10; // -> 10°C
    }

    Scalar referencePressureAtPos(const GlobalPosition& globalPos) const
    {
        return 1.0e5; // -> 1 bar
    }
    /*!
     * \brief DOC ME!
     * \param values DOC ME!
     * \param globalPos DOC ME!
     */
    void sourceAtPos(PrimaryVariables &values,
                     const GlobalPosition& globalPos) const
    {
        values = 0;
    }
    /*!
     * \brief DOC ME!
     * \param bcTypes DOC ME!
     * \param globalPos DOC ME!
     */
    void boundaryTypesAtPos(BoundaryTypes &bcTypes,
                            const GlobalPosition& globalPos) const
    {
        //west
        if (globalPos[0] < eps_)
            bcTypes.setAllDirichlet();

        // the other boundary (east)
        else
            bcTypes.setAllNeumann();
    }
    /*!
     * \brief DOC ME!
     * \param values DOC ME!
     * \param globalPos DOC ME!
     */
    void dirichletAtPos(PrimaryVariables &values,
                        const GlobalPosition& globalPos) const
    {
        values[pnIdx] = pLeftBc_;
        values[swIdx] = 1.0 - snr_;
    }
    /*!
     * \brief DOC ME!
     * \param values DOC ME!
     * \param globalPos DOC ME!
     */
    void neumannAtPos(PrimaryVariables &values,
                      const GlobalPosition& globalPos) const
    {
        values = 0; //homogeneous Neumann
    }
    /*!
     * \brief DOC ME!
     * \param values DOC ME!
     * \param globalPos DOC ME!
     */
    void initialAtPos(PrimaryVariables &values,
                      const GlobalPosition &globalPos) const
    {
        values[pnIdx] = pLeftBc_;
        values[swIdx] = swr_;
    }

    //Override outputfunction for ViPLab-Output
    void writeOutput()
    {
        if (paraviewOutput_)
            ParentType::writeOutput(false);

        else
        {
            std::size_t numElements = this->gridView().size(0);
            viplabOutput_.writeHeader();
            std::vector<int> color({0, 0, 255}); // blue
            std::vector<Scalar> solutionVectorSaturation(numElements);
            for (std::size_t i = 0; i < numElements; i++)
                solutionVectorSaturation[i] = this->variables().cellData(i).saturation(wPhaseIdx);

            viplabOutput_.writeVipLabOutput(solutionVectorSaturation, color);

            for (std::size_t i = 0; i < numElements; i++)
                solutionVectorSaturation[i]=analyticSolution_.AnalyticSolution()[i];

            color = {255, 0, 0}; // red
            viplabOutput_.writeVipLabOutput(solutionVectorSaturation, color);
            viplabOutput_.writeLegend();
        }
    }

private:
    Scalar eps_, swr_, snr_;
    Scalar pLeftBc_;
    McWhorterAnalytic<TypeTag> analyticSolution_;
    ViplabOutput<TypeTag> viplabOutput_;
    bool paraviewOutput_;

};

} // end namespace Dumux

#endif
